# -*- coding: utf-8 -*-
"""
Base Model Unittests
====================
Modified: 2022-03
"""

import unittest
import logging
from unittest.mock import patch, MagicMock

from myosin.models.base import BaseModel


class TestBaseModel(unittest.TestCase):

    @patch.multiple(
        BaseModel,
        __abstractmethods__=set(),
        serialize=MagicMock(),
        deserialize=MagicMock()
    )
    def setUp(self) -> None:
        logging.disable()
        # test set id
        self.base = BaseModel(1)  # type: ignore
        # test unset id (autogeneration)
        self.comparator = BaseModel()  # type: ignore

    def tearDown(self) -> None:
        del self.base
        logging.disable(logging.NOTSET)

    def test_hash(self):
        """
        Test hashing
        """
        self.assertNotEqual(self.base.__hash__(), self.comparator.__hash__())

    def test_typehash(self):
        """
        Test type hashing
        """
        self.assertEqual(self.base.__typehash__(), self.comparator.__typehash__())

    def test_repr(self):
        """
        Test base model repr
        """
        self.assertEqual(type(self.base.__repr__()), str)

    def test_eq(self):
        """
        Test base model equality comparator
        """
        # test base model comparisons with different ids
        self.assertFalse(self.base == self.comparator)
        # test base model comparison with same id
        self.comparator.id = 1
        self.assertTrue(self.base == self.comparator)
        # test non model comparison
        self.assertFalse(self.base == object())

    def test_id(self):
        """
        Test base model id get/set
        """
        self.base.id = 2
        self.assertEqual(self.base.id, 2)



# -*- coding: utf-8 -*-
"""
State Unittests
===============
Modified: 2022-04

"""

import unittest
import logging
from unittest.mock import patch
from tests.resources.models import DemoState

from myosin import State
from myosin.state.ssm import SSM
from myosin.exceptions.state import HashNotFound, NullCheckoutError, UninitializedStateError


class TestState(unittest.TestCase):

    def setUp(self) -> None:
        logging.disable()
        self.test_state = DemoState(1)
        with State() as state:
            self.state = state

    def tearDown(self) -> None:
        logging.disable(logging.NOTSET)
        self.state.reset()

    def mock_loader(self):
        """
        State loading helper to remove interdependancy of testing methods
        """
        self.state._ssm[self.test_state.__typehash__()] = SSM[DemoState](self.test_state)

    def test_uninitialized_load(self):
        """
        Test loading an uninitalized model
        """
        with self.assertRaises(UninitializedStateError):
            self.state.load(self.test_state)

    def test_load(self):
        """
        Test state load keyset and SSM wrapper configuration
        """
        # initialize model
        self.test_state.name = "test"
        self.state.load(self.test_state)
        # im setting the test like this in order to test both the keyset and the object set,
        # I want to raise test assertion failures rather than key errors.
        ssm = self.state._ssm.get(self.test_state.__typehash__())
        self.assertIsNotNone(ssm)
        if ssm is not None:
            self.assertEqual(ssm.ref, self.test_state)

    def test_null_checkout(self):
        """
        Test null checkout fails with NullCheckoutError
        """
        with self.assertRaises(NullCheckoutError):
            _ = self.state.checkout(DemoState)

    def test_checkout(self):
        """
        Test checkout copy
        """
        self.mock_loader()
        test_state = self.state.checkout(DemoState)
        self.assertNotEqual(hash(test_state), hash(self.test_state))

    def test_unregistered_commit(self):
        """
        Test unregistered state commit
        """
        with self.assertRaises(HashNotFound):
            self.state.commit(self.test_state)

    @patch('asyncio.run', lambda x: None)
    def test_commit(self):
        """
        Test state commit
        """
        self.mock_loader()
        self.state.commit(self.test_state)



# -*- coding: utf-8 -*-
"""
Myosin State Engine
===================
Modified: 2022-04

"""
import asyncio
import copy
import logging
from typing import Dict, Type, TypeVar, Callable

from myosin.state.ssm import SSM
from myosin.typing import AsyncCallback
from myosin.utils.funcs import pformat
from myosin.models.state import StateModel
from myosin.utils.concurrency import ThreadUtils as tutils
from myosin.exceptions.state import HashNotFound, NullCheckoutError, UninitializedStateError

# generic runtime model type
_T = TypeVar('_T', bound=StateModel)


class State:

    # storage mechanism is { name: Model }
    _ssm: Dict[int, SSM] = {}

    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb): ...

    def load(self, model: _T) -> _T:
        """
        Load state model into memory

        :param model: state model
        :type model: StateModel
        """
        self._ssm[model.__typehash__()] = SSM[_T](model)
        # assert the model attributes are initialized
        try:
            model.serialize()
        except AttributeError as exc:
            raise UninitializedStateError from exc
        self._logger.info("Loaded state model: %s", pformat(model.serialize()))
        return model

    @tutils.lock(tutils.state_lock)
    def checkout(self, state_type: Type[_T]) -> _T:
        """
        Returns reference of requested state model

        :param state_type: requested state model reference 
        :type state_type: Type[_T]
        :raises NullCheckoutError: if the requested state type does not exist 
        :return: deep copy of requested state model
        :rtype: _T
        """
        self._logger.info("Checking out state model of type %s", state_type)
        _type_hash = hash(state_type)
        self._logger.debug("Computed type hash: %s", _type_hash)
        ssm = self._ssm.get(_type_hash)
        if not ssm:
            raise NullCheckoutError
        return copy.deepcopy(ssm.ref)

    @tutils.lock(tutils.state_lock)
    def commit(self, state: _T, cache: bool = False) -> _T:
        """
        Commit new state to system state and update state subscriber callbacks

        :param state: modified copy of state
        :type state: _T
        :param cache: cache the state to disk once updated, defaults to False
        :type cache: bool, optional
        :raises HashNotFound: if system state has no state registered of the requested type 
        :return: updated system state reference
        :rtype: _T
        """
        # automatic type inference by typehash
        _type_hash = hash(type(state))
        # verify typehash
        ssm = self._ssm.get(_type_hash)
        if not ssm:
            self._logger.error("Committed typehash: %s did not match any state model", _type_hash)
            raise HashNotFound
        ssm.ref = state
        asyncio.run(ssm.execute())
        if cache:
            state.cache()
            self._logger.debug("Cached commited state model %s", state)
        return state

    def subscribe(self, state_type: Type[_T], callback: Callable[[_T], AsyncCallback]) -> None:
        """
        Subscribe an asynchronous state change listener to a designated runtime model

        :param state_type: model type to subscribe to
        :type state_type: Type[T]
        :param callback: state change listener callback
        :type callback: Callable[[T], AsyncCallback]
        """
        _type_hash = hash(state_type)
        ssm = self._ssm.get(_type_hash)
        if not ssm:
            self._logger.error("Subscribed typehash: %s did not match any state model", _type_hash)
            raise HashNotFound
        ssm.queue.append(callback)

    def reset(self) -> None:
        """
        Reset all loaded state model
        """
        self._ssm.clear()



# -*- coding: utf-8 -*-
"""
Myosin
======
Modified: 2022-03

Lightweight & threadsafe state engine
"""

import logging

from myosin.state import State
from myosin.__version__ import __version__
from myosin.models.base import BaseModel
from myosin.models.state import StateModel

__all__ = [
    '__version__',
    'BaseModel',
    'StateModel',
    'State'
]

_log = logging.getLogger(__name__)
_log.info("Myosin version %s", __version__)



# -*- coding: utf-8 -*-
"""
Concurrency Utils Unittests
===========================
Modified: 2022-03
"""

import unittest
import logging

from myosin.utils.concurrency import ThreadUtils as tutils


class TestConcurrency(unittest.TestCase):

    def setUp(self) -> None:
        logging.disable()

    def tearDown(self) -> None:
        logging.disable(logging.NOTSET)

    def test_lock(self):
        """
        Test lock decorator
        """
        @tutils.lock(tutils.state_lock)
        def test_method() -> bool:
            self.assertTrue(tutils.state_lock.locked())
            return True
        self.assertFalse(tutils.state_lock.locked())
        self.assertTrue(test_method())



# -*- coding: utf-8 -*-
"""
SSM Unittests
=============
Modified: 2022-04
"""

import unittest
import logging
import asyncio
from unittest.mock import AsyncMock
from myosin.state.ssm import SSM
from tests.resources.models import DemoState


class TestSSM(unittest.TestCase):

    def setUp(self) -> None:
        logging.disable()
        self.test_state = DemoState(1)
        self.ssm = SSM[DemoState](self.test_state)

    def tearDown(self) -> None:
        del self.ssm
        logging.disable(logging.NOTSET)

    def test_reference_hash(self):
        """
        Test for correct reference hash return
        """
        print(hash(self.ssm.ref))
        self.assertEqual(self.ssm.refhash, hash(self.test_state))

    def test_type_hash(self):
        """
        Test for correct type hash return
        """
        self.assertEqual(self.ssm.typehash, self.test_state.__typehash__())

    def test_ref(self):
        """
        Test reference object get/set
        """
        new_state = DemoState(2)
        self.ssm.ref = new_state
        self.assertEqual(self.ssm.ref, new_state)

    def test_queue(self):
        """
        Test async queue callback get/set
        """
        async def async_callback(_: DemoState): ...
        new_queue = [async_callback]
        self.ssm.queue = new_queue
        self.assertEqual(self.ssm.queue, new_queue)

    def test_execute(self):
        """
        Test execution runner and exception reporting
        """
        alpha_cb = AsyncMock()
        alpha_cb.side_effect = BaseException
        beta_cb = AsyncMock()
        self.ssm.queue = [alpha_cb, beta_cb]
        asyncio.run(self.ssm.execute())
        alpha_cb.assert_called_once()
        beta_cb.assert_called_once()



import os
import uuid
import logging
from threading import Thread
import time
from typing import Any, Dict
from datetime import datetime

from myosin import StateModel
from myosin import State


class User(StateModel):

    def __init__(self, _id: int) -> None:
        super().__init__(_id)

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name

    @property
    def email(self) -> str:
        return self.__email

    @email.setter
    def email(self, email: str) -> None:
        self.__email = email

    def serialize(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'name': self.name,
            'email': self.email
        }

    def deserialize(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)


class Telemetry(StateModel):

    def __init__(self) -> None:
        super().__init__()

    @property
    def tp(self) -> float:
        return self.__tp

    @tp.setter
    def tp(self, value: float) -> None:
        self.__tp = value

    @property
    def timestamp(self) -> float:
        return datetime.now().timestamp()

    def serialize(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'temp': self.tp,
            'timestamp': self.timestamp
        }

    def deserialize(self, **kwargs) -> None:
        for k, v in kwargs.items():
            if k == "timestamp":
                continue
            setattr(self, k, v)


class Runner:
    def __init__(self) -> None:
        self.username = ""
        with State() as state:
            state.subscribe(User, self.update)

    async def update(self, usr: User) -> None:
        self.username = usr.name

    def run(self):
        while True:
            time.sleep(0.001)
            print(f"This is my username: {self.username}")


if __name__ == "__main__":
    os.environ['MYOSIN_CACHE_BASE_PATH'] = "tests/cache"
    logging.basicConfig(level=logging.DEBUG)
    _log = logging.getLogger(__name__)

    u = User(_id=1)
    u.deserialize(**{'name': "chris", 'email': "chris@email.io"})
    t = Telemetry()
    t.deserialize(**{'tp': 65.0})
    u.email = "bad"
    _log.info(u)

    with State() as state:
        state.load(u)
        state.load(t)

    with State() as state:
        user = state.checkout(User)
        _log.info(user.email)

    with State() as state:
        user = state.checkout(User)
        user.email = "bad@email.com"
        state.commit(user, cache=True)

    with State() as state:
        user = state.checkout(User)
        _log.info(user)

    with State() as state:
        telemetry = state.checkout(Telemetry)
        _log.info(telemetry)

    with State() as state:
        telemetry = state.checkout(Telemetry)
        telemetry.tp = 45.0
        state.commit(telemetry, cache=True)

    # test state model subscriptions
    _runner = Runner()
    Thread(target=_runner.run, args=(), daemon=True).start()
    while True:
        time.sleep(0.001)
        with State() as state:
            usr = state.checkout(User)
            usr.name = str(uuid.uuid4())
            state.commit(usr)




import json
from typing import Any


def pformat(payload: Any) -> str:
    """
    Format payloads for logging

    :param payload: payload to format
    :type payload: Any
    :return: logging output
    :rtype: str
    """
    return "\n" + json.dumps(payload, indent=2)
